<div class="fflex">
	<sly-select-agent 
        ref="sly-select-agent-{{{ widget.widget_id }}}"
	    :value="state.{{{ widget.widget_id }}}.agentId" 
        :group-id="data.{{{ widget.widget_id }}}.teamId" 
        :options="state.{{{ widget.widget_id }}}.options"
	    :is-community="data.{{{ widget.widget_id }}}.isCommunity" 
        {% if widget._changes_handled == true %} 
            @input="state.{{{ widget.widget_id }}}.agentId = $event; post('/{{{ widget.widget_id }}}/value_changed')"
        {% else %}
            @input="state.{{{ widget.widget_id }}}.agentId = $event"
        {% endif %} 
    ></sly-select-agent>

	<el-tooltip :open-delay="300" :enterable="false" content="Refresh">
		<el-button
			type="text"
			class="ml5"
			style="padding: 0; font-size: 20px"
			@click="$refs['sly-select-agent-{{{ widget.widget_id }}}'].updateNodes()"
			><i class="zmdi zmdi-refresh mr5"></i></el-button
	></el-tooltip>
</div>

