.color-picker-compact .el-color-picker .el-color-picker__trigger {
    height: 0px;
    padding: 0;
    border: none;
}

.color-picker-compact .el-color-picker .el-color-picker__icon {
    display: inline-block;
    position: relative;
    top: -6px;
    margin-left: -16.5px;
    width: 12px;
    color: #888;
    font-size: 12px;
}

.color-picker-compact .el-color-picker .el-color-picker__trigger .el-color-picker__color {
    position: relative;
    display: inline-block;
    box-sizing: border-box;
    border: 1px solid #666;
    width: 22px;
    height: 22px;
    text-align: center;
    border-radius: 2px
}

.color-picker-compact .el-color-picker .el-color-picker__trigger .el-color-picker__icon.el-icon-caret-bottom {
    color: white;
}